from .base import Column


class TupleColumn(Column):
    py_types = (list, tuple)

    def __init__(self, nested_columns, **kwargs):
        self.nested_columns = nested_columns
        super(TupleColumn, self).__init__(**kwargs)

    async def write_data(
        self, items,
    ):
        if not items:
            return
        items = list(zip(*items))

        for i, x in enumerate(self.nested_columns):
            await x.write_data(list(items[i]),)

    async def write_items(
        self, items,
    ):
        return await self.write_data(items,)

    async def read_data(
        self, n_items,
    ):
        rv = [await x.read_data(n_items,) for x in self.nested_columns]
        return list(zip(*rv))

    async def read_items(
        self, n_items,
    ):
        return await self.read_data(n_items,)


def create_tuple_column(spec, column_by_spec_getter, column_options):
    brackets = 0
    column_begin = 0

    inner_spec = get_inner_spec(spec)
    nested_columns = []
    for i, x in enumerate(inner_spec + ","):
        if x == ",":
            if brackets == 0:
                nested_columns.append(inner_spec[column_begin:i])
                column_begin = i + 1
        elif x == "(":
            brackets += 1
        elif x == ")":
            brackets -= 1
        elif x == " ":
            if brackets == 0:
                column_begin = i + 1

    return TupleColumn([column_by_spec_getter(x) for x in nested_columns], **column_options)


def get_inner_spec(spec):
    brackets = 1
    offset = len("Tuple(")
    i = offset
    for i, ch in enumerate(spec[offset:], offset):
        if brackets == 0:
            break

        if ch == "(":
            brackets += 1

        elif ch == ")":
            brackets -= 1

    return spec[offset:i]



from .lz4 import Compressor as BaseCompressor
from .lz4 import Decompressor as BaseDecompressor


class Compressor(BaseCompressor):
    mode = "high_compression"


class Decompressor(BaseDecompressor):
    pass



from asynch.proto.io import BufferedReader, BufferedWriter


class BlockInfo:
    is_overflows = False
    bucket_num = -1

    def __init__(self, writer: BufferedWriter = None, reader: BufferedReader = None):
        self.writer = writer
        self.reader = reader

    async def write(self):
        # Set of pairs (`FIELD_NUM`, value) in binary form. Then 0.
        await self.writer.write_varint(1)
        await self.writer.write_uint8(self.is_overflows)
        await self.writer.write_varint(2)
        await self.writer.write_int32(self.bucket_num)
        await self.writer.write_varint(0)

    async def read(self):
        while True:
            field_num = await self.reader.read_varint()
            if not field_num:
                break

            if field_num == 1:
                self.is_overflows = bool(await self.reader.read_uint8())

            elif field_num == 2:
                self.bucket_num = await self.reader.read_uint32()


class BaseBlock:
    def __init__(
        self,
        writer: BufferedWriter = None,
        reader: BufferedReader = None,
        columns_with_types=None,
        data=None,
        info=None,
        types_check=False,
    ):
        self.writer = writer
        self.reader = reader
        self.columns_with_types = columns_with_types or []
        self.types_check = types_check
        self.info = info or BlockInfo(writer, reader)
        self.data = self.normalize(data or [])

    def normalize(self, data):
        return data

    @property
    def num_columns(self):
        raise NotImplementedError

    @property
    def num_rows(self):
        raise NotImplementedError

    def get_columns(self):
        raise NotImplementedError

    def get_rows(self):
        raise NotImplementedError

    def get_column_by_index(self, index):
        raise NotImplementedError

    def transposed(self):
        return list(zip(*self.data))


class ColumnOrientedBlock(BaseBlock):
    def normalize(self, data):
        if not data:
            return []

        self._check_number_of_columns(data)
        self._check_all_columns_equal_length(data)
        return data

    @property
    def num_columns(self):
        return len(self.data)

    @property
    def num_rows(self):
        return len(self.data[0]) if self.num_columns else 0

    def get_columns(self):
        return self.data

    def get_rows(self):
        return self.transposed()

    def get_column_by_index(self, index):
        return self.data[index]

    def _check_number_of_columns(self, data):
        expected_row_len = len(self.columns_with_types)

        got = len(data)
        if expected_row_len != got:
            msg = "Expected {} columns, got {}".format(expected_row_len, got)
            raise ValueError(msg)

    def _check_all_columns_equal_length(self, data):
        expected = len(data[0])

        for column in data:
            got = len(column)
            if got != expected:
                msg = "Expected {} rows, got {}".format(expected, got)
                raise ValueError(msg)


class RowOrientedBlock(BaseBlock):
    dict_row_types = (dict,)
    tuple_row_types = (list, tuple)
    supported_row_types = dict_row_types + tuple_row_types

    def normalize(self, data):
        if not data:
            return []

        # Guessing about whole data format by first row.
        first_row = data[0]

        if self.types_check:
            self._check_row_type(first_row)

        if isinstance(first_row, dict):
            self._mutate_dicts_to_rows(data)
        else:
            self._check_rows(data)

        return data

    @property
    def num_columns(self):
        return len(self.data[0]) if self.num_rows else 0

    @property
    def num_rows(self):
        return len(self.data)

    def get_columns(self):
        return self.transposed()

    def get_rows(self):
        return self.data

    def get_column_by_index(self, index):
        return [row[index] for row in self.data]

    def _mutate_dicts_to_rows(self, data):
        column_names = [x[0] for x in self.columns_with_types]

        check_row_type = False
        if self.types_check:
            check_row_type = self._check_dict_row_type

        for i, row in enumerate(data):
            if check_row_type:
                check_row_type(row)

            data[i] = [row[name] for name in column_names]

    def _check_rows(self, data):
        expected_row_len = len(self.columns_with_types)

        got = len(data[0])
        if expected_row_len != got:
            msg = "Expected {} columns, got {}".format(expected_row_len, got)
            raise ValueError(msg)

        if self.types_check:
            check_row_type = self._check_tuple_row_type
            for row in data:
                check_row_type(row)

    def _check_row_type(self, row):
        if not isinstance(row, self.supported_row_types):
            raise TypeError(
                "Unsupported row type: {}. dict, list or tuple is expected.".format(type(row))
            )

    def _check_tuple_row_type(self, row):
        if not isinstance(row, self.tuple_row_types):
            raise TypeError(
                "Unsupported row type: {}. list or tuple is expected.".format(type(row))
            )

    def _check_dict_row_type(self, row):
        if not isinstance(row, self.dict_row_types):
            raise TypeError("Unsupported row type: {}. dict is expected.".format(type(row)))


class BlockStreamProfileInfo:
    def __init__(self, reader: BufferedReader):
        self.rows = 0
        self.blocks = 0
        self.bytes = 0
        self.applied_limit = False  # bool
        self.rows_before_limit = 0
        self.calculated_rows_before_limit = 0  # bool
        self.reader = reader

    async def read(self,):
        self.rows = await self.reader.read_varint()
        self.blocks = await self.reader.read_varint()
        self.bytes = await self.reader.read_varint()
        self.applied_limit = bool(await self.reader.read_uint8())
        self.rows_before_limit = await self.reader.read_varint()
        self.calculated_rows_before_limit = bool(await self.reader.read_uint8())



import ssl
from typing import Type
from urllib.parse import parse_qs, unquote, urlparse

from asynch import errors
from asynch.cursors import Cursor
from asynch.proto.connection import Connection as ProtoConnection
from asynch.proto.utils.compat import asbool


class Connection:
    def __init__(
        self,
        dsn: str = None,
        host: str = "127.0.0.1",
        port: int = 9000,
        database: str = "default",
        user: str = "default",
        password: str = "",
        cursor_cls=Cursor,
        echo=False,
        stack_track=False,
        **kwargs,
    ):
        self._dsn = dsn
        self._user = user
        self._password = password
        self._host = host
        self._port = port
        self._database = database
        self._connection_kwargs = kwargs
        self._is_closed = False
        self._echo = echo
        self._cursor_cls = cursor_cls
        self._connected = False
        if dsn:
            self._connection = ProtoConnection(
                **self._parse_dsn(dsn), stack_track=stack_track, **kwargs
            )
        else:
            self._connection = ProtoConnection(
                host=host,
                port=port,
                database=database,
                user=user,
                password=password,
                stack_track=stack_track,
                **kwargs,
            )

    def __repr__(self):
        return "<connection object at 0x{0:x}; closed: {1:}>".format(id(self), self._is_closed)

    @property
    def connected(self):
        return self._connected

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def user(self):
        return self._user

    @property
    def password(self):
        return self._password

    @property
    def database(self):
        return self._database

    @property
    def echo(self):
        return self._echo

    async def close(self):
        if self._is_closed:
            return
        await self._connection.disconnect()
        self._is_closed = True

    async def commit(self):
        raise errors.NotSupportedError

    async def rollback(self):
        raise errors.NotSupportedError

    async def connect(self):
        if self._connected:
            return
        await self._connection.connect()
        self._connected = True

    def cursor(self, cursor: Type[Cursor] = None) -> Cursor:
        cursor_cls = cursor or self._cursor_cls
        return cursor_cls(self, self._echo)

    def _parse_dsn(self, url):
        """
        Return a client configured from the given URL.

        For example::

            clickhouse://[user:password]@localhost:9000/default
            clickhouses://[user:password]@localhost:9440/default

        Three URL schemes are supported:
            clickhouse:// creates a normal TCP socket connection
            clickhouses:// creates a SSL wrapped TCP socket connection

        Any additional querystring arguments will be passed along to
        the Connection class's initializer.
        """
        url = urlparse(url)

        settings = {}
        kwargs = {"host": url.hostname}

        if url.port is not None:
            self._port = kwargs["port"] = url.port

        path = url.path.replace("/", "", 1)
        if path:
            self._database = kwargs["database"] = path

        if url.username is not None:
            self._user = kwargs["user"] = unquote(url.username)

        if url.password is not None:
            self._password = kwargs["password"] = unquote(url.password)

        if url.scheme == "clickhouses":
            kwargs["secure"] = True

        compression_algs = {"lz4", "lz4hc", "zstd"}
        timeouts = {"connect_timeout", "send_receive_timeout", "sync_request_timeout"}

        for name, value in parse_qs(url.query).items():
            if not value or not len(value):
                continue

            value = value[0]

            if name == "compression":
                value = value.lower()
                if value in compression_algs:
                    kwargs[name] = value
                else:
                    kwargs[name] = asbool(value)

            elif name == "secure":
                kwargs[name] = asbool(value)

            elif name == "client_name":
                kwargs[name] = value

            elif name in timeouts:
                kwargs[name] = float(value)

            elif name == "compress_block_size":
                kwargs[name] = int(value)

            # ssl
            elif name == "verify":
                kwargs[name] = asbool(value)
            elif name == "ssl_version":
                kwargs[name] = getattr(ssl, value)
            elif name in ["ca_certs", "ciphers"]:
                kwargs[name] = value
            elif name == "alt_hosts":
                kwargs["alt_hosts"] = value
            else:
                settings[name] = value

        if settings:
            kwargs["settings"] = settings

        return kwargs


async def connect(
    dsn: str = None,
    host: str = "127.0.0.1",
    port: int = 9000,
    database: str = "default",
    user: str = "default",
    password: str = "",
    cursor_cls=Cursor,
    echo=False,
    **kwargs,
) -> Connection:
    conn = Connection(dsn, host, port, database, user, password, cursor_cls, echo=echo, **kwargs)
    await conn.connect()
    return conn



from asynch.proto.io import BufferedWriter
from asynch.proto.utils.compat import asbool


class SettingType:
    @classmethod
    async def write(cls, writer: BufferedWriter, value):
        raise NotImplementedError


class SettingUInt64(SettingType):
    @classmethod
    async def write(cls, writer, value):
        await writer.write_varint(int(value))


class SettingBool(SettingType):
    @classmethod
    async def write(cls, writer, value):
        await writer.write_varint(asbool(value))


class SettingString(SettingType):
    @classmethod
    async def write(cls, writer, value):
        await writer.write_str(value)


class SettingChar(SettingType):
    @classmethod
    async def write(cls, writer, value):
        await writer.write_str(value[0],)


class SettingFloat(SettingType):
    @classmethod
    async def write(cls, writer, value):
        """
        Float is written in string representation.
        """
        await writer.write_str(str(value),)


class SettingMaxThreads(SettingUInt64):
    @classmethod
    async def write(cls, writer, value):
        if value == "auto":
            value = 0
        await super(SettingMaxThreads, cls).write(writer, value)



from ...errors import (
    ColumnTypeMismatchException,
    StructPackException,
    TypeMismatchError,
    UnknownTypeError,
)
from ..io import BufferedReader, BufferedWriter
from .arraycolumn import create_array_column
from .datecolumn import DateColumn
from .datetimecolumn import create_datetime_column
from .decimalcolumn import create_decimal_column
from .enumcolumn import create_enum_column
from .floatcolumn import Float32, Float64
from .intcolumn import (
    Int8Column,
    Int16Column,
    Int32Column,
    Int64Column,
    UInt8Column,
    UInt16Column,
    UInt32Column,
    UInt64Column,
)
from .intervalcolumn import (
    IntervalDayColumn,
    IntervalHourColumn,
    IntervalMinuteColumn,
    IntervalMonthColumn,
    IntervalSecondColumn,
    IntervalWeekColumn,
    IntervalYearColumn,
)
from .ipcolumn import IPv4Column, IPv6Column
from .lowcardinalitycolumn import create_low_cardinality_column
from .nothingcolumn import NothingColumn
from .nullablecolumn import create_nullable_column
from .nullcolumn import NullColumn
from .simpleaggregatefunctioncolumn import create_simple_aggregate_function_column
from .stringcolumn import create_string_column
from .tuplecolumn import create_tuple_column
from .uuidcolumn import UUIDColumn

column_by_type = {
    c.ch_type: c
    for c in [
        DateColumn,
        Float32,
        Float64,
        Int8Column,
        Int16Column,
        Int32Column,
        Int64Column,
        UInt8Column,
        UInt16Column,
        UInt32Column,
        UInt64Column,
        NothingColumn,
        NullColumn,
        UUIDColumn,
        IntervalYearColumn,
        IntervalMonthColumn,
        IntervalWeekColumn,
        IntervalDayColumn,
        IntervalHourColumn,
        IntervalMinuteColumn,
        IntervalSecondColumn,
        IPv4Column,
        IPv6Column,
    ]
}


def get_column_by_spec(spec, column_options):
    def create_column_with_options(x):
        return get_column_by_spec(x, column_options)

    if spec == "String" or spec.startswith("FixedString"):
        return create_string_column(spec, column_options)

    elif spec.startswith("Enum"):
        return create_enum_column(spec, column_options)

    elif spec.startswith("DateTime"):
        return create_datetime_column(spec, column_options)

    elif spec.startswith("Decimal"):
        return create_decimal_column(spec, column_options)

    elif spec.startswith("Array"):
        return create_array_column(spec, create_column_with_options)

    elif spec.startswith("Tuple"):
        return create_tuple_column(spec, create_column_with_options, column_options)

    elif spec.startswith("Nullable"):
        return create_nullable_column(spec, create_column_with_options)

    elif spec.startswith("LowCardinality"):
        return create_low_cardinality_column(spec, create_column_with_options)

    elif spec.startswith("SimpleAggregateFunction"):
        return create_simple_aggregate_function_column(spec, create_column_with_options)

    else:
        try:
            cls = column_by_type[spec]
            return cls(**column_options)

        except KeyError as e:
            raise UnknownTypeError("Unknown type {}".format(e.args[0]))


async def read_column(
    reader: BufferedReader, writer: BufferedWriter, context, column_spec, n_items,
):
    column_options = {"context": context, "reader": reader, "writer": writer}
    column = get_column_by_spec(column_spec, column_options)
    await column.read_state_prefix()
    return await column.read_data(n_items,)


async def write_column(
    reader: BufferedReader,
    writer: BufferedWriter,
    context,
    column_name,
    column_spec,
    items,
    types_check=False,
):
    column_options = {
        "context": context,
        "types_check": types_check,
        "reader": reader,
        "writer": writer,
    }
    column = get_column_by_spec(column_spec, column_options)

    try:
        await column.write_state_prefix()
        await column.write_data(items)

    except ColumnTypeMismatchException as e:
        raise TypeMismatchError(
            "Type mismatch in VALUES section. "
            'Expected {} got {}: {} for column "{}".'.format(
                column_spec, type(e.args[0]), e.args[0], column_name
            )
        )

    except (StructPackException, OverflowError) as e:
        error = e.args[0]
        raise TypeMismatchError(
            "Type mismatch in VALUES section. "
            "Repeat query with types_check=True for detailed info. "
            "Column {}: {}".format(column_name, str(error))
        )



from urllib.parse import parse_qs, unquote, urlparse  # noqa: F401

string_types = (str,)
integer_types = (int,)
class_types = (type,)
text_type = str
binary_type = bytes
range = range
StandardError = Exception


# from paste.deploy.converters
def asbool(obj):
    if isinstance(obj, string_types):
        obj = obj.strip().lower()
        if obj in ["true", "yes", "on", "y", "t", "1"]:
            return True
        elif obj in ["false", "no", "off", "n", "f", "0"]:
            return False
        else:
            raise ValueError("String is not true/false: %r" % obj)
    return bool(obj)



from uuid import UUID

from ...errors import CannotParseUuidError
from ..io import MAX_UINT64
from ..utils import compat
from .base import FormatColumn


class UUIDColumn(FormatColumn):
    ch_type = "UUID"
    py_types = compat.string_types + (UUID,)
    format = "Q"

    # UUID is stored by two uint64 numbers.
    async def write_items(
        self, items,
    ):
        n_items = len(items)

        uint_64_pairs = [None] * 2 * n_items
        for i, x in enumerate(items):
            i2 = 2 * i
            uint_64_pairs[i2] = (x >> 64) & MAX_UINT64
            uint_64_pairs[i2 + 1] = x & MAX_UINT64

        s = self.make_struct(2 * n_items)
        await self.writer.write_bytes(s.pack(*uint_64_pairs))

    async def read_items(
        self, n_items,
    ):
        # TODO: cythonize
        s = self.make_struct(2 * n_items)
        items = s.unpack(await self.reader.read_bytes(s.size))

        uint_128_items = [None] * n_items
        for i in range(n_items):
            i2 = 2 * i
            uint_128_items[i] = (items[i2] << 64) + items[i2 + 1]

        return tuple(uint_128_items)

    def after_read_items(self, items, nulls_map=None):
        if nulls_map is None:
            return tuple(UUID(int=item) for item in items)
        else:
            return tuple(
                (None if is_null else UUID(int=items[i])) for i, is_null in enumerate(nulls_map)
            )

    def before_write_items(self, items, nulls_map=None):
        null_value = self.null_value

        for i, item in enumerate(items):
            if nulls_map and nulls_map[i]:
                items[i] = null_value
                continue

            try:
                if not isinstance(item, UUID):
                    item = UUID(item)

            except ValueError:
                raise CannotParseUuidError("Cannot parse uuid '{}'".format(item))

            items[i] = item.int



import asyncio
import logging
from time import time
from types import GeneratorType
from typing import AsyncGenerator, Optional
from urllib.parse import urlparse

from asynch.errors import (
    ServerException,
    UnexpectedPacketFromServerError,
    UnknownPacketFromServerError,
)
from asynch.proto import constants
from asynch.proto.block import BlockStreamProfileInfo, ColumnOrientedBlock, RowOrientedBlock
from asynch.proto.compression import get_compressor_cls
from asynch.proto.context import Context
from asynch.proto.cs import ClientInfo, QueryKind, ServerInfo
from asynch.proto.io import BufferedReader, BufferedWriter
from asynch.proto.progress import Progress
from asynch.proto.protocol import ClientPacket, Compression, ServerPacket
from asynch.proto.result import IterQueryResult, ProgressQueryResult, QueryInfo, QueryResult
from asynch.proto.settings import write_settings
from asynch.proto.streams.native import BlockInputStream, BlockOutputStream
from asynch.proto.utils.escape import escape_params
from asynch.proto.utils.helpers import chunks, column_chunks

logger = logging.getLogger(__name__)


class QueryProcessingStage:
    """
    Determines till which state SELECT query should be executed.
    """

    FETCH_COLUMNS = 0
    WITH_MERGEABLE_STATE = 1
    COMPLETE = 2


class Packet:
    def __init__(self):
        self.type = None
        self.block = None
        self.exception = None
        self.progress = None
        self.profile_info = None
        self.multistring_message = None


class Connection:
    log_priorities = (
        "Unknown",
        "Fatal",
        "Critical",
        "Error",
        "Warning",
        "Notice",
        "Information",
        "Debug",
        "Trace",
    )

    def __init__(  # nosec:B107
        self,
        host: str = "127.0.0.1",
        port: int = 9000,
        database: str = "default",
        user: str = "default",
        password: str = "",
        client_name: str = constants.CLIENT_NAME,
        connect_timeout: int = constants.DBMS_DEFAULT_CONNECT_TIMEOUT_SEC,
        send_receive_timeout: int = constants.DBMS_DEFAULT_TIMEOUT_SEC,
        sync_request_timeout: int = constants.DBMS_DEFAULT_SYNC_REQUEST_TIMEOUT_SEC,
        compress_block_size: int = constants.DEFAULT_COMPRESS_BLOCK_SIZE,
        compression: bool = False,
        secure: bool = False,
        # Secure socket parameters.
        verify: bool = True,
        ssl_version=None,
        ca_certs=None,
        ciphers=None,
        alt_hosts: str = None,
        stack_track=False,
        **kwargs,
    ):
        self.stack_track = stack_track
        if secure:
            default_port = constants.DEFAULT_SECURE_PORT
        else:
            default_port = constants.DEFAULT_PORT
        self.hosts = [(host, port or default_port)]
        if alt_hosts:
            for host in alt_hosts.split(","):
                url = urlparse("clickhouse://" + host)
                self.hosts.append((url.hostname, url.port or default_port))
        self.database = database
        self.user = user
        self.password = password
        self.client_name = constants.DBMS_NAME + " " + client_name
        self.connect_timeout = connect_timeout
        self.send_receive_timeout = send_receive_timeout
        self.sync_request_timeout = sync_request_timeout
        self.secure_socket = secure
        self.verify_cert = verify

        ssl_options = {}
        if ssl_version is not None:
            ssl_options["ssl_version"] = ssl_version
        if ca_certs is not None:
            ssl_options["ca_certs"] = ca_certs
        if ciphers is not None:
            ssl_options["ciphers"] = ciphers
        self.ssl_options = ssl_options
        # Use LZ4 compression by default.
        if compression is True:
            compression = "lz4"

        if compression is False:
            self.compression = Compression.DISABLED
            self.compressor_cls = None
            self.compress_block_size = None
        else:
            self.compression = Compression.ENABLED
            self.compressor_cls = get_compressor_cls(compression)
            self.compress_block_size = compress_block_size
        self.connected = False
        self.reader: Optional[BufferedReader] = None
        self.writer: Optional[BufferedWriter] = None
        self.server_info: Optional[ServerInfo] = None
        self.context = Context()
        # Block writer/reader
        self.block_in_stream: Optional[BlockInputStream] = None
        self.block_out_stream: Optional[BlockOutputStream] = None
        self.settings = kwargs.pop("settings", {}).copy()
        self.client_settings = {
            "insert_block_size": int(
                self.settings.pop("insert_block_size", constants.DEFAULT_INSERT_BLOCK_SIZE,)
            ),
            "strings_as_bytes": self.settings.pop("strings_as_bytes", False),
            "strings_encoding": self.settings.pop("strings_encoding", constants.STRINGS_ENCODING),
        }
        self.last_query: Optional[QueryInfo] = None
        self.available_client_settings = (
            "insert_block_size",  # TODO: rename to max_insert_block_size
            "strings_as_bytes",
            "strings_encoding",
        )
        self.context.settings = self.settings
        self.context.client_settings = self.client_settings

    def get_block_in_stream(self):
        if self.compression:
            from asynch.proto.streams.compressed import CompressedBlockInputStream

            return CompressedBlockInputStream(self.reader, self.writer, self.context)
        else:
            from asynch.proto.streams.native import BlockInputStream

            return BlockInputStream(self.reader, self.writer, self.context)

    def get_block_out_stream(self):
        if self.compression:
            from .streams.compressed import CompressedBlockOutputStream

            return CompressedBlockOutputStream(
                self.reader,
                self.writer,
                self.compressor_cls,
                self.compress_block_size,
                self.context,
            )
        else:
            from .streams.native import BlockOutputStream

            return BlockOutputStream(self.reader, self.writer, self.context)

    async def send_hello(self):
        await self.writer.write_varint(ClientPacket.HELLO)
        await self.writer.write_str(self.client_name)
        await self.writer.write_varint(constants.CLIENT_VERSION_MAJOR)
        await self.writer.write_varint(constants.CLIENT_VERSION_MINOR)
        await self.writer.write_varint(constants.CLIENT_REVISION)
        await self.writer.write_str(self.database)
        await self.writer.write_str(self.user)
        await self.writer.write_str(self.password)
        await self.writer.flush()

    def get_server(self):
        return "{}:{}".format(self.host, self.port)

    def unexpected_packet_message(self, expected, packet_type):
        packet_type = ServerPacket.to_str(packet_type)

        return "Unexpected packet from server {} (expected {}, got {})".format(
            self.get_server(), expected, packet_type
        )

    async def receive_hello(self):
        packet_type = await self.reader.read_varint()
        if packet_type == ServerPacket.HELLO:
            server_name = await self.reader.read_str()
            server_version_major = await self.reader.read_varint()
            server_version_minor = await self.reader.read_varint()
            server_revision = await self.reader.read_varint()
            server_timezone = None
            if server_revision >= constants.DBMS_MIN_REVISION_WITH_SERVER_TIMEZONE:
                server_timezone = await self.reader.read_str()

            server_display_name = ""
            if server_revision >= constants.DBMS_MIN_REVISION_WITH_SERVER_DISPLAY_NAME:
                server_display_name = await self.reader.read_str()

            server_version_patch = server_revision
            if server_revision >= constants.DBMS_MIN_REVISION_WITH_VERSION_PATCH:
                server_version_patch = await self.reader.read_varint()

            self.server_info = ServerInfo(
                server_name,
                server_version_major,
                server_version_minor,
                server_version_patch,
                server_revision,
                server_timezone,
                server_display_name,
            )
            self.context.server_info = self.server_info

            logger.debug(
                "Connected to %s server version %s.%s.%s, revision: %s",
                server_name,
                server_version_major,
                server_version_minor,
                server_version_patch,
                server_revision,
            )
        elif packet_type == ServerPacket.EXCEPTION:
            raise await self.read_exception()
        else:
            await self.disconnect()
            message = self.unexpected_packet_message("Hello or Exception", packet_type)
            raise UnexpectedPacketFromServerError(message)

    async def ping(self):
        await self.writer.write_varint(ClientPacket.PING)
        await self.writer.flush()
        packet_type = await self.reader.read_varint()
        while packet_type == ServerPacket.PROGRESS:
            await self.receive_progress()
            packet_type = await self.reader.read_varint()
        if packet_type != ServerPacket.PONG:
            msg = self.unexpected_packet_message("Pong", packet_type)
            raise UnexpectedPacketFromServerError(msg)
        return True

    async def receive_data(self):
        revision = self.server_info.revision

        if revision >= constants.DBMS_MIN_REVISION_WITH_TEMPORARY_TABLES:
            await self.reader.read_str()

        return await self.block_in_stream.read()

    async def receive_exception(self):
        return await self.read_exception()

    async def read_exception(self, additional_message=None):
        code = await self.reader.read_int32()
        name = await self.reader.read_str()
        message = await self.reader.read_str()
        stack_trace = await self.reader.read_str()
        has_nested = bool(await self.reader.read_uint8())

        new_message = ""

        if additional_message:
            new_message += additional_message + ". "

        if name != "DB::Exception":
            new_message += name + ". "

        new_message += message
        if self.stack_track:
            new_message += ". Stack trace:\n\n" + stack_trace
        nested = None
        if has_nested:
            nested = await self.read_exception()

        return ServerException(new_message, code, nested=nested)

    async def receive_profile_info(self):
        profile_info = BlockStreamProfileInfo(self.reader)
        await profile_info.read()
        return profile_info

    async def receive_multistring_message(self, packet_type: int):
        num = ServerPacket.strings_in_message(packet_type)
        return [await self.reader.read_str() for _i in range(num)]

    def log_block(self, block):
        column_names = [x[0] for x in block.columns_with_types]

        for row in block.get_rows():
            row = dict(zip(column_names, row))

            if 1 <= row["priority"] <= 8:
                priority = self.log_priorities[row["priority"]]
            else:
                priority = row[0]

            # thread_number in servers prior 20.x
            thread_id = row.get("thread_id") or row["thread_number"]

            logger.info(
                "[ %s ] [ %s ] {%s} <%s> %s: %s",
                row["host_name"],
                thread_id,
                row["query_id"],
                priority,
                row["source"],
                row["text"],
            )

    async def receive_packet(self):
        packet = await self._receive_packet()

        if packet.type == ServerPacket.EXCEPTION:
            raise packet.exception

        elif packet.type == ServerPacket.PROGRESS:
            self.last_query.store_progress(packet.progress)
            return packet

        elif packet.type == ServerPacket.END_OF_STREAM:
            return False

        elif packet.type == ServerPacket.DATA:
            return packet

        elif packet.type == ServerPacket.TOTALS:
            return packet

        elif packet.type == ServerPacket.EXTREMES:
            return packet

        elif packet.type == ServerPacket.PROFILE_INFO:
            self.last_query.store_profile(packet.profile_info)
            return True

        else:
            return True

    async def _receive_packet(self):
        packet = Packet()

        packet.type = packet_type = await self.reader.read_varint()

        if packet_type == ServerPacket.DATA:
            packet.block = await self.receive_data()

        elif packet_type == ServerPacket.EXCEPTION:
            packet.exception = await self.receive_exception()

        elif packet.type == ServerPacket.PROGRESS:
            packet.progress = await self.receive_progress()
        elif packet.type == ServerPacket.PROFILE_INFO:
            packet.profile_info = await self.receive_profile_info()
        elif packet_type == ServerPacket.TOTALS:
            packet.block = await self.receive_data()

        elif packet_type == ServerPacket.EXTREMES:
            packet.block = await self.receive_data()

        elif packet_type == ServerPacket.LOG:
            block = await self.receive_data()
            self.log_block(block)
        elif packet_type == ServerPacket.END_OF_STREAM:
            pass

        elif packet_type == ServerPacket.TABLE_COLUMNS:
            packet.multistring_message = await self.receive_multistring_message(packet_type)

        else:
            await self.disconnect()
            raise UnknownPacketFromServerError(
                "Unknown packet {} from server {}".format(packet_type, self.get_server())
            )

        return packet

    async def packet_generator(self) -> AsyncGenerator:
        while True:
            packet = await self.receive_packet()
            if not packet:
                break

            if packet is True:
                continue

            yield packet

    async def receive_result(self, with_column_types=False, progress=False, columnar=False):

        generator = self.packet_generator()

        if progress:
            return ProgressQueryResult(
                self.reader, generator, with_column_types=with_column_types, columnar=columnar
            )

        else:
            result = QueryResult(
                self.reader, generator, with_column_types=with_column_types, columnar=columnar
            )
            return await result.get_result()

    async def receive_progress(self):
        progress = Progress(self.reader)
        await progress.read(self.server_info.revision,)
        return progress

    async def send_cancel(self):
        await self.writer.write_varint(ClientPacket.CANCEL)
        await self.writer.flush()

    async def send_query(self, query: str, query_id: str = ""):
        await self.writer.write_varint(ClientPacket.QUERY)
        await self.writer.write_str(query_id)
        revision = self.server_info.revision
        if revision >= constants.DBMS_MIN_REVISION_WITH_CLIENT_INFO:
            client_info = ClientInfo(self.client_name)
            client_info.query_kind = QueryKind.INITIAL_QUERY
            await client_info.write(revision, self.writer)

        settings_as_strings = (
            revision >= constants.DBMS_MIN_REVISION_WITH_SETTINGS_SERIALIZED_AS_STRINGS
        )
        await write_settings(self.writer, self.context.settings, settings_as_strings)

        await self.writer.write_varint(QueryProcessingStage.COMPLETE)
        await self.writer.write_varint(self.compression)

        await self.writer.write_str(query,)

        logger.debug("Query: %s", query)

        await self.writer.flush()

    async def _init_connection(self, host: str, port: int):
        self.host, self.port = host, port
        reader, writer = await asyncio.open_connection(host, port)
        self.writer = BufferedWriter(writer, constants.BUFFER_SIZE)
        self.reader = BufferedReader(reader, constants.BUFFER_SIZE)
        self.block_in_stream = self.get_block_in_stream()
        self.block_out_stream = self.get_block_out_stream()
        self.connected = True
        await self.send_hello()
        await self.receive_hello()

    async def disconnect(self):
        if self.connected:
            self.connected = False
            await self.writer.close()

    async def connect(self):
        if self.connected:
            await self.disconnect()
        logger.debug("Connecting. Database: %s. User: %s", self.database, self.user)
        for host, port in self.hosts:
            logger.debug("Connecting to %s:%s", host, port)
            return await self._init_connection(host, port)

    async def execute(
        self,
        query,
        args=None,
        with_column_types=False,
        external_tables=None,
        query_id="",
        settings=None,
        types_check=False,
        columnar=False,
    ):
        """
        Executes query.

        Establishes new connection if it wasn't established yet.
        After query execution connection remains intact for next queries.
        If connection can't be reused it will be closed and new connection will
        be created.

        :param query: query that will be send to server.
        :param args: substitution parameters for SELECT queries and data for
                       INSERT queries. Data for INSERT can be `list`, `tuple`
                       or :data:`~types.GeneratorType`.
                       Defaults to ``None`` (no parameters  or data).
        :param with_column_types: if specified column names and types will be
                                  returned alongside with result.
                                  Defaults to ``False``.
        :param external_tables: external tables to send.
                                Defaults to ``None`` (no external tables).
        :param query_id: the query identifier. If no query id specified
                         ClickHouse server will generate it.
        :param settings: dictionary of query settings.
                         Defaults to ``None`` (no additional settings).
        :param types_check: enables type checking of data for INSERT queries.
                            Causes additional overhead. Defaults to ``False``.
        :param columnar: if specified the result of the SELECT query will be
                         returned in column-oriented form.
                         It also allows to INSERT data in columnar form.
                         Defaults to ``False`` (row-like form).

        :return: * number of inserted rows for INSERT queries with data.
                   Returning rows count from INSERT FROM SELECT is not
                   supported.
                 * if `with_column_types=False`: `list` of `tuples` with
                   rows/columns.
                 * if `with_column_types=True`: `tuple` of 2 elements:
                    * The first element is `list` of `tuples` with
                      rows/columns.
                    * The second element information is about columns: names
                      and types.
        """

        start_time = time()
        self.make_query_settings(settings)
        await self.force_connect()
        self.last_query = QueryInfo(self.reader)

        # INSERT queries can use list/tuple/generator of list/tuples/dicts.
        # For SELECT parameters can be passed in only in dict right now.
        is_insert = isinstance(args, (list, tuple, GeneratorType))

        if is_insert:
            rv = await self.process_insert_query(
                query,
                args,
                external_tables=external_tables,
                query_id=query_id,
                types_check=types_check,
                columnar=columnar,
            )
        else:
            rv = await self.process_ordinary_query(
                query,
                args,
                with_column_types=with_column_types,
                external_tables=external_tables,
                query_id=query_id,
                types_check=types_check,
                columnar=columnar,
            )
        self.last_query.store_elapsed(time() - start_time)
        return rv

    def make_query_settings(self, settings):
        settings = dict(settings or {})

        # Pick client-related settings.
        client_settings = self.client_settings.copy()
        for key in self.available_client_settings:
            if key in settings:
                client_settings[key] = settings.pop(key)

        self.context.client_settings = client_settings

        # The rest of settings are ClickHouse-related.
        query_settings = self.settings.copy()
        query_settings.update(settings)
        self.context.settings = query_settings

    async def force_connect(self):
        if not self.connected:
            await self.connect()

        elif not await self.ping():
            logger.warning("Connection was closed, reconnecting.")
            await self.connect()

    async def process_ordinary_query(
        self,
        query,
        params=None,
        with_column_types=False,
        external_tables=None,
        query_id="",
        types_check=False,
        columnar=False,
    ):

        if params is not None:
            query = self.substitute_params(query, params)

        await self.send_query(query, query_id=query_id)
        await self.send_external_tables(external_tables, types_check=types_check)
        return await self.receive_result(with_column_types=with_column_types, columnar=columnar)

    async def send_external_tables(self, tables, types_check=False):
        for table in tables or []:
            block = RowOrientedBlock(
                self.writer, self.reader, table["structure"], table["data"], types_check=types_check
            )
            await self.send_block(block, table_name=table["name"])

        # Empty block, end of data transfer.
        await self.send_block(RowOrientedBlock(self.writer, self.reader))

    async def send_block(self, block, table_name=""):
        await self.writer.write_varint(ClientPacket.DATA)

        revision = self.server_info.revision
        if revision >= constants.DBMS_MIN_REVISION_WITH_TEMPORARY_TABLES:
            await self.writer.write_str(table_name,)

        await self.block_out_stream.write(block)

    def substitute_params(self, query, params):
        if not isinstance(params, dict):
            raise ValueError("Parameters are expected in dict form")

        escaped = escape_params(params)
        return query % escaped

    async def process_insert_query(
        self,
        query_without_data,
        data,
        external_tables=None,
        query_id=None,
        types_check=False,
        columnar=False,
    ):
        await self.send_query(query_without_data, query_id=query_id)
        await self.send_external_tables(external_tables, types_check=types_check)

        sample_block = await self.receive_sample_block()
        if sample_block:
            rv = await self.send_data(
                sample_block, data, types_check=types_check, columnar=columnar
            )
            packet = await self._receive_packet()
            if packet.exception:
                raise packet.exception
            return rv

    async def receive_sample_block(self):
        while True:
            packet = await self._receive_packet()

            if packet.type == ServerPacket.DATA:
                return packet.block

            elif packet.type == ServerPacket.EXCEPTION:
                raise packet.exception

            elif packet.type == ServerPacket.TABLE_COLUMNS:
                pass

            else:
                message = self.unexpected_packet_message(
                    "Data, Exception or TableColumns", packet.type
                )
                raise UnexpectedPacketFromServerError(message)

    async def send_data(self, sample_block, data, types_check=False, columnar=False):
        inserted_rows = 0

        client_settings = self.context.client_settings
        block_cls = ColumnOrientedBlock if columnar else RowOrientedBlock
        slicer = column_chunks if columnar else chunks

        for chunk in slicer(data, client_settings["insert_block_size"]):
            block = block_cls(
                self.writer,
                self.reader,
                columns_with_types=sample_block.columns_with_types,
                data=chunk,
                types_check=types_check,
            )
            await self.send_block(block)
            inserted_rows += block.num_rows

        # Empty block means end of data.
        await self.send_block(block_cls(self.writer, self.reader))
        return inserted_rows

    async def iter_process_ordinary_query(
        self,
        query,
        params=None,
        with_column_types=False,
        external_tables=None,
        query_id=None,
        types_check=False,
    ):

        if params is not None:
            query = self.substitute_params(query, params)

        await self.send_query(query, query_id=query_id)
        await self.send_external_tables(external_tables, types_check=types_check)
        return self.iter_receive_result(with_column_types=with_column_types)

    def iter_receive_result(self, with_column_types=False):
        gen = self.packet_generator()

        for rows in IterQueryResult(gen, with_column_types=with_column_types):
            for row in rows:
                yield row

    async def execute_iter(
        self,
        query,
        params=None,
        with_column_types=False,
        external_tables=None,
        query_id=None,
        settings=None,
        types_check=False,
    ):
        """
        *New in version 0.0.14.*

        Executes SELECT query with results streaming. See, :ref:`execute-iter`.

        :param query: query that will be send to server.
        :param params: substitution parameters for SELECT queries and data for
                       INSERT queries. Data for INSERT can be `list`, `tuple`
                       or :data:`~types.GeneratorType`.
                       Defaults to ``None`` (no parameters  or data).
        :param with_column_types: if specified column names and types will be
                                  returned alongside with result.
                                  Defaults to ``False``.
        :param external_tables: external tables to send.
                                Defaults to ``None`` (no external tables).
        :param query_id: the query identifier. If no query id specified
                         ClickHouse server will generate it.
        :param settings: dictionary of query settings.
                         Defaults to ``None`` (no additional settings).
        :param types_check: enables type checking of data for INSERT queries.
                            Causes additional overhead. Defaults to ``False``.
        :return: :ref:`iter-query-result` proxy.
        """

        self.make_query_settings(settings)
        await self.force_connect()
        self.last_query = QueryInfo(self.reader)

        return await self.iter_process_ordinary_query(
            query,
            params=params,
            with_column_types=with_column_types,
            external_tables=external_tables,
            query_id=query_id,
            types_check=types_check,
        )



from asynch.proto import constants
from asynch.proto.io import BufferedReader


class Progress:
    def __init__(self, reader: BufferedReader):
        self.rows = 0
        self.bytes = 0
        self.total_rows = 0
        self.written_rows = 0
        self.written_bytes = 0
        self.reader = reader

    async def read(
        self, server_revision,
    ):
        self.rows = await self.reader.read_varint()
        self.bytes = await self.reader.read_varint()

        revision = server_revision
        if revision >= constants.DBMS_MIN_REVISION_WITH_TOTAL_ROWS_IN_PROGRESS:
            self.total_rows = await self.reader.read_varint()

        if revision >= constants.DBMS_MIN_REVISION_WITH_CLIENT_WRITE_INFO:
            self.written_rows = await self.reader.read_varint()
            self.written_bytes = await self.reader.read_varint()

    def increment(self, another_progress):
        self.rows += another_progress.rows
        self.bytes += another_progress.bytes
        self.total_rows += another_progress.total_rows
        self.written_rows += another_progress.written_rows
        self.written_bytes += another_progress.written_bytes



import importlib
from typing import Callable

from clickhouse_cityhash.cityhash import CityHash128

from asynch.errors import ChecksumDoesntMatchError, UnknownCompressionMethod
from asynch.proto.io import BufferedReader, BufferedWriter
from asynch.proto.protocol import CompressionMethodByte


def get_compressor_cls(alg):
    try:
        module = importlib.import_module("." + alg, __name__)
        return module.Compressor

    except ImportError:
        raise UnknownCompressionMethod("Unknown compression method: '{}'".format(alg))


def get_decompressor_cls(method_type):
    if method_type == CompressionMethodByte.LZ4:
        module = importlib.import_module(".lz4", __name__)

    elif method_type == CompressionMethodByte.ZSTD:
        module = importlib.import_module(".zstd", __name__)

    else:
        raise UnknownCompressionMethod()

    return module.Decompressor


class BaseCompressor:
    """
    Partial file-like object with write method.
    """

    method = None
    method_byte = None
    compress_func: Callable

    def __init__(self, writer: BufferedWriter):
        self.writer = writer

    def get_value(self):
        value = self.writer.buffer
        return value

    def write(self, p_str):
        self.writer.write_str(p_str)

    async def get_compressed_data(self, extra_header_size):
        data = self.get_value()
        compressed = self.compress_func(data)
        header_size = extra_header_size + 4 + 4  # sizes
        await self.writer.write_uint32(header_size + len(compressed))
        await self.writer.write_uint32(len(data))
        await self.writer.write_str(compressed)
        return self.writer.buffer


class BaseDecompressor:
    method = None
    method_byte = None
    decompress_func: Callable

    def __init__(self, reader: BufferedReader):
        self.reader = reader

    def check_hash(self, compressed_data, compressed_hash):
        if CityHash128(compressed_data) != compressed_hash:
            raise ChecksumDoesntMatchError()

    async def get_decompressed_data(self, method_byte, compressed_hash, extra_header_size):
        size_with_header = await self.reader.read_uint32()
        compressed_size = size_with_header - extra_header_size - 4

        compressed = await self.reader.read_bytes(compressed_size)

        writer = BufferedWriter()
        await writer.write_uint8(method_byte)
        await writer.write_uint32(size_with_header)
        await writer.write_bytes(compressed.getvalue())
        self.check_hash(writer.buffer, compressed_hash)
        compressed = compressed.read(compressed_size - 4)
        return compressed



import asyncio
import collections
from asyncio import Condition
from collections.abc import Coroutine
from typing import Deque, Set

from asynch.connection import Connection, connect


class _ContextManager(Coroutine):
    __slots__ = ("_coro", "_obj")

    def __init__(self, coro):
        self._coro = coro
        self._obj = None

    def send(self, value):
        return self._coro.send(value)

    def throw(self, typ, val=None, tb=None):
        if val is None:
            return self._coro.throw(typ)
        elif tb is None:
            return self._coro.throw(typ, val)
        else:
            return self._coro.throw(typ, val, tb)

    def close(self):
        return self._coro.close()

    @property
    def gi_frame(self):
        return self._coro.gi_frame

    @property
    def gi_running(self):
        return self._coro.gi_running

    @property
    def gi_code(self):
        return self._coro.gi_code

    def __next__(self):
        return self.send(None)

    def __iter__(self):
        return self._coro.__await__()

    def __await__(self):
        return self._coro.__await__()

    async def __aenter__(self):
        self._obj = await self._coro
        return self._obj

    async def __aexit__(self, exc_type, exc, tb):
        await self._obj.close()
        self._obj = None


class _PoolContextManager(_ContextManager):
    async def __aexit__(self, exc_type, exc, tb):
        self._obj.close()
        await self._obj.wait_closed()
        self._obj = None


class _PoolAcquireContextManager(_ContextManager):
    __slots__ = ("_coro", "_conn", "_pool")

    def __init__(self, coro, pool):
        super().__init__(coro)
        self._coro = coro
        self._conn = None
        self._pool = pool

    async def __aenter__(self):
        self._conn = await self._coro
        return self._conn

    async def __aexit__(self, exc_type, exc, tb):
        try:
            await self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None


class Pool(asyncio.AbstractServer):
    def __init__(self, minsize: int = 1, maxsize: int = 10, loop=None, **kwargs):
        self._maxsize = maxsize
        self._minsize = minsize
        self._connection_kwargs = kwargs
        self._terminated: Set[Connection] = set()
        self._used: Set[Connection] = set()
        self._cond = Condition()
        self._closing = False
        self._closed = False
        self._free: Deque[Connection] = collections.deque(maxlen=maxsize)
        self._loop = loop

        if maxsize <= 0:
            raise ValueError("maxsize is expected to be greater than zero")

        if minsize < 0:
            raise ValueError("minsize is expected to be greater or equal to zero")

        if minsize > maxsize:
            raise ValueError("minsize is greater than max_size")

    @property
    def maxsize(self):
        return self._maxsize

    @property
    def minsize(self):
        return self._minsize

    @property
    def freesize(self):
        return len(self._free)

    @property
    def size(self):
        return self.freesize + len(self._used)

    @property
    def cond(self):
        return self._cond

    async def release(self, connection: Connection):
        """Release free connection back to the connection pool.

        This is **NOT** a coroutine.
        """
        fut = self._loop.create_future()
        fut.set_result(None)

        if connection in self._terminated:
            self._terminated.remove(connection)
            return fut
        self._used.remove(connection)
        if connection.connected:
            if self._closing:
                await connection.close()
            else:
                self._free.append(connection)
            fut = self._loop.create_task(self._wakeup())
        return fut

    async def _wakeup(self):
        async with self._cond:
            self._cond.notify()

    def _wait(self):
        return len(self._terminated) > 0

    def acquire(self):
        return _PoolAcquireContextManager(self._acquire(), self)

    async def _acquire(self) -> Connection:
        if self._closing:
            raise RuntimeError("Cannot acquire connection after closing pool")
        async with self._cond:
            while True:
                await self.initialize()
                if self._free:
                    conn = self._free.popleft()
                    self._used.add(conn)
                    return conn
                else:
                    await self._cond.wait()

    async def initialize(self):
        while self.size < self.minsize:
            conn = await connect(**self._connection_kwargs)
            self._free.append(conn)
            self._cond.notify()

    async def clear(self):
        """Close all free connections in pool."""
        async with self._cond:
            while self._free:
                conn = self._free.popleft()
                await conn.close()
            self._cond.notify()

    async def wait_closed(self):
        """Wait for closing all pool's connections."""

        if self._closed:
            return
        if not self._closing:
            raise RuntimeError(".wait_closed() should be called " "after .close()")

        while self._free:
            conn = self._free.popleft()
            await conn.close()

        async with self._cond:
            while self.size > self.freesize:
                await self._cond.wait()

        self._closed = True

    def close(self):
        """Close pool.

        Mark all pool connections to be closed on getting back to pool.
        Closed pool doesn't allow to acquire new connections.
        """
        if self._closed:
            return
        self._closing = True

    async def terminate(self):
        """Terminate pool.

        Close pool with instantly closing all acquired connections also.
        """

        self.close()

        for conn in self._used:
            await conn.close()
            self._terminated.add(conn)

        self._used.clear()


def create_pool(minsize: int = 1, maxsize: int = 10, loop=None, **kwargs):
    coro = _create_pool(minsize=minsize, maxsize=maxsize, loop=loop, **kwargs)
    return _PoolContextManager(coro)


async def _create_pool(minsize: int = 1, maxsize: int = 10, loop=None, **kwargs):
    if loop is None:
        loop = asyncio.get_event_loop()
    pool = Pool(minsize, maxsize, loop, **kwargs)
    if minsize > 0:
        async with pool.cond:
            await pool.initialize()
    return pool



import zstd

from asynch.proto.compression import BaseCompressor, BaseDecompressor
from asynch.proto.protocol import CompressionMethod, CompressionMethodByte


class Compressor(BaseCompressor):
    method = CompressionMethod.ZSTD
    method_byte = CompressionMethodByte.ZSTD
    compress_func = zstd.compress


class Decompressor(BaseDecompressor):
    method = CompressionMethod.ZSTD
    method_byte = CompressionMethodByte.ZSTD
    decompress_func = zstd.decompress

    async def get_decompressed_data(self, method_byte, compressed_hash, extra_header_size):
        compressed = await super(Decompressor, self).get_decompressed_data(
            method_byte, compressed_hash, extra_header_size
        )

        return zstd.decompress(compressed)



from asynch.proto.columns.base import Column
from asynch.proto.io import BufferedReader, BufferedWriter
from asynch.proto.utils import compat


class String(Column):
    ch_type = "String"
    py_types = compat.string_types
    null_value = ""

    async def write_items(
        self, items,
    ):
        await self.writer.write_strings(items)

    async def read_items(
        self, n_items,
    ):
        ret = []
        for _ in range(n_items):
            ret.append(await self.reader.read_str())
        return tuple(ret)


class ByteString(String):
    py_types = (bytes,)
    null_value = b""


class FixedString(String):
    ch_type = "FixedString"

    def __init__(self, reader: BufferedReader, writer: BufferedWriter, length: int, **kwargs):
        self.length = length
        super().__init__(reader, writer, **kwargs)


class ByteFixedString(FixedString):
    py_types = (bytearray, bytes)
    null_value = b""

    async def write_items(
        self, items,
    ):
        for item in items:
            await self.writer.write_bytes(item)


def create_string_column(spec, column_options):
    client_settings = column_options["context"].client_settings
    strings_as_bytes = client_settings["strings_as_bytes"]
    if spec == "String":
        cls = ByteString if strings_as_bytes else String
        return cls(**column_options)
    else:
        length = int(spec[12:-1])
        cls = ByteFixedString if strings_as_bytes else FixedString
        return cls(length=length, **column_options)



import logging
from collections import namedtuple
from itertools import islice

from asynch.errors import InterfaceError, ProgrammingError

Column = namedtuple("Column", "name type_code display_size internal_size precision scale null_ok")

logger = logging.getLogger(__name__)


class States:
    (NONE, RUNNING, FINISHED, CURSOR_CLOSED) = range(4)


class Cursor:
    _states = States()
    _columns_with_types = None

    def __init__(self, connection=None, echo=False):
        self._connection = connection
        self._reset_state()
        self._rows = []
        self._echo = echo
        self._arraysize = 1

    @property
    def connection(self):
        """This read-only attribute return a reference to the Connection
        object on which the cursor was created."""
        return self._connection

    @property
    def rowcount(self):
        """
        :return: the number of rows that the last .execute*() produced.
        """
        return self._rowcount

    def setinputsizes(self, *args):
        """Does nothing, required by DB API."""

    def setoutputsizes(self, *args):
        """Does nothing, required by DB API."""

    async def close(self):
        conn = self._connection
        if conn is None:
            return
        await self._connection.close()
        self._connection = None

    async def execute(
        self, query: str, args=None,
    ):
        self._check_cursor_closed()
        self._begin_query()

        execute, execute_kwargs = self._prepare()

        response = await execute(query, args=args, with_column_types=True, **execute_kwargs)

        self._process_response(response)
        self._end_query()
        if self._echo:
            logger.info(query)
            logger.info("%r", args)
        return self._rowcount

    def _process_response(self, response, executemany=False):
        if executemany or isinstance(response, int):
            self._rowcount = response
            response = None

        if not response:
            self._columns = self._types = self._rows = []
            return

        if self._stream_results:
            columns_with_types = next(response)
            rows = response

        else:
            rows, columns_with_types = response
        self._columns_with_types = columns_with_types
        if columns_with_types:
            self._columns, self._types = zip(*columns_with_types)
            if not self._stream_results:
                self._rowcount = len(rows)
        else:
            self._columns = self._types = []

        self._rows = rows

    async def executemany(self, query, args=None):
        self._check_cursor_closed()
        self._begin_query()

        execute, execute_kwargs = self._prepare()

        response = await execute(query, args=args, **execute_kwargs)

        self._process_response(response, executemany=True)
        self._end_query()
        if self._echo:
            logger.info(query)
            logger.info("%r", args)
        return self._rowcount

    def fetchone(self):
        self._check_query_started()

        if self._stream_results:
            return next(self._rows, None)

        else:
            if not self._rows:
                return None

            return self._rows.pop(0)

    def fetchmany(self, size: int):
        self._check_query_started()

        if size is None:
            size = self._arraysize

        if self._stream_results:
            if size == -1:
                return list(self._rows)
            else:
                return list(islice(self._rows, size))

        if size < 0:
            rv = self._rows
            self._rows = []
        else:
            rv = self._rows[:size]
            self._rows = self._rows[size:]

        return rv

    def fetchall(self,):
        self._check_query_started()

        if self._stream_results:
            return list(self._rows)

        rv = self._rows
        self._rows = []
        return rv

    def _reset_state(self):
        """
        Resets query state and get ready for another query.
        """
        self._state = self._states.NONE

        self._columns = None
        self._types = None
        self._rows = None
        self._rowcount = -1

        self._stream_results = False
        self._max_row_buffer = 0
        self._settings = None
        self._query_id = ""
        self._external_tables = {}
        self._types_check = False

    def _prepare(self):
        external_tables = [
            {"name": name, "structure": structure, "data": data}
            for name, (structure, data) in self._external_tables.items()
        ] or None

        execute = self._connection._connection.execute

        if self._stream_results:
            execute = self._connection._connection.execute_iter
            self.settings = self.settings or {}
            self.settings["max_block_size"] = self._max_row_buffer

        execute_kwargs = {
            "settings": self._settings,
            "external_tables": external_tables,
            "types_check": self._types_check,
            "query_id": self._query_id,
        }

        return execute, execute_kwargs

    def __iter__(self):
        while True:
            one = self.fetchone()
            if one is None:
                return
            yield one

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def __aenter__(self):
        return self

    @property
    def description(self):
        if self._state == self._states.NONE:
            return None

        columns = self._columns or []
        types = self._types or []

        return [
            Column(name, type_code, None, None, None, None, True)
            for name, type_code in zip(columns, types)
        ]

    def _check_query_started(self):
        if self._state == self._states.NONE:
            raise ProgrammingError("no results to fetch")

    def _check_cursor_closed(self):
        if self._state == self._states.CURSOR_CLOSED:
            raise InterfaceError("cursor already closed")

    def _begin_query(self):
        self._state = self._states.RUNNING

    def _end_query(self):
        self._state = self._states.FINISHED

    def set_stream_results(self, stream_results, max_row_buffer):
        """
        Toggles results streaming from server. Driver will consume
        block-by-block of `max_row_buffer` size and yield row-by-row from each
        block.

        :param stream_results: enable or disable results streaming.
        :param max_row_buffer: specifies the maximum number of rows to buffer
               at a time.
        :return: None
        """
        self._stream_results = stream_results
        self._max_row_buffer = max_row_buffer

    def set_settings(self, settings):
        """
        Specifies settings for cursor.

        :param settings: dictionary of query settings
        :return: None
        """
        self._settings = settings

    def set_types_check(self, types_check):
        """
        Toggles type checking for sequence of INSERT parameters.
        Disabled by default.

        :param types_check: new types check value.
        :return: None
        """
        self._types_check = types_check

    def set_external_table(self, name, structure, data):
        """
        Adds external table to cursor context.

        If the same table is specified more than once the last one is used.

        :param name: name of external table
        :param structure: list of tuples (name, type) that defines table
                          structure. Example [(x, 'Int32')].
        :param data: sequence of rows of tuples or dicts for transmission.
        :return: None
        """
        self._external_tables[name] = (structure, data)

    def set_query_id(self, query_id=""):
        """
        Specifies the query identifier for cursor.

        :param query_id: the query identifier.
        :return: None
        """
        self._query_id = query_id

    # End non-PEP methods


class DictCursor(Cursor):
    def _process_response(self, response, executemany=False):
        super(DictCursor, self)._process_response(response, executemany)
        if self._columns and self._rows:
            self._rows = [dict(zip(self._columns, item)) for item in self._rows]



import logging
from typing import Dict

from asynch.proto.io import BufferedWriter
from asynch.proto.settings.available import settings as available_settings

logger = logging.getLogger(__name__)


async def write_settings(writer: BufferedWriter, settings: Dict, settings_as_strings):
    is_important = 0

    for setting, value in (settings or {}).items():
        # If the server support settings as string we do not need to know
        # anything about them, so we can write any setting.
        if settings_as_strings:
            await writer.write_str(setting)
            await writer.write_uint8(is_important,)
            await writer.write_str(str(value))

        else:
            # If the server requires string in binary,
            # then they cannot be written without type.
            setting_writer = available_settings.get(setting)
            if not setting_writer:
                logger.warning("Unknown setting %s. Skipping", setting)
                continue
            await writer.write_str(setting,)
            await setting_writer.write(writer, value)

    await writer.write_str("")  # end of settings



from typing import Optional

from asynch.proto.cs import ServerInfo


class Context:
    def __init__(self):
        self._server_info: Optional[ServerInfo] = None
        self._settings = {}
        self._client_settings = {}

    @property
    def server_info(self):
        return self._server_info

    @server_info.setter
    def server_info(self, value):
        self._server_info = value

    @property
    def settings(self):
        return self._settings.copy()

    @settings.setter
    def settings(self, value):
        self._settings = value.copy()

    @property
    def client_settings(self):
        return self._client_settings.copy()

    @client_settings.setter
    def client_settings(self, value):
        self._client_settings = value.copy()



from math import log

from .base import Column
from .intcolumn import UInt8Column, UInt16Column, UInt32Column, UInt64Column


def create_low_cardinality_column(spec, column_by_spec_getter):
    inner = spec[15:-1]
    nested = column_by_spec_getter(inner)
    return LowCardinalityColumn(nested)


class LowCardinalityColumn(Column):
    """
    Stores column as index (unique elements) and keys.
    Good for de-duplication of large values with low cardinality.
    """

    int_types = {0: UInt8Column, 1: UInt16Column, 2: UInt32Column, 3: UInt64Column}

    # Need to read additional keys.
    # Additional keys are stored before indexes as value N and N keys
    # after them.
    has_additional_keys_bit = 1 << 9
    # Need to update dictionary.
    # It means that previous granule has different dictionary.
    need_update_dictionary = 1 << 10

    serialization_type = has_additional_keys_bit | need_update_dictionary

    def __init__(self, nested_column, **kwargs):
        kwargs.update(dict(reader=nested_column.reader, writer=nested_column.writer))
        self.nested_column = nested_column
        super(LowCardinalityColumn, self).__init__(**kwargs)

    async def read_state_prefix(self,):
        return await self.reader.read_uint64()

    async def write_state_prefix(self,):
        # KeysSerializationVersion. See ClickHouse docs.
        return await self.writer.write_int64(1)

    async def _write_data(
        self, items,
    ):
        index, keys = [], []
        key_by_index_element = {}

        if self.nested_column.nullable:
            # First element represents NULL if column is nullable.
            index.append(self.nested_column.null_value)
            # Prevent null map writing. Reset nested column nullable flag.
            self.nested_column.nullable = False

            for x in items:
                if x is None:
                    # Zero element for null.
                    keys.append(0)

                else:
                    key = key_by_index_element.get(x)
                    # Get key from index or add it to index.
                    if key is None:
                        key = len(key_by_index_element)
                        key_by_index_element[x] = key
                        index.append(x)

                    keys.append(key + 1)
        else:
            for x in items:
                key = key_by_index_element.get(x)

                # Get key from index or add it to index.
                if key is None:
                    key = len(key_by_index_element)
                    key_by_index_element[x] = len(key_by_index_element)
                    index.append(x)

                keys.append(key)

        # Do not write anything for empty column.
        # May happen while writing empty arrays.
        if not len(index):
            return

        int_type = int(log(len(index), 2) / 8)
        int_column = self.int_types[int_type](reader=self.reader, writer=self.writer)

        serialization_type = self.serialization_type | int_type
        await self.writer.write_int64(serialization_type)
        await self.writer.write_int64(len(index))

        await self.nested_column.write_data(index,)
        await self.writer.write_int64(len(items))

        await int_column.write_data(keys,)

    async def _read_data(self, n_items, nulls_map=None):
        if not n_items:
            return tuple()

        serialization_type = await self.reader.read_uint64()

        # Lowest byte contains info about key type.
        key_type = serialization_type & 0xF
        keys_column = self.int_types[key_type]()

        nullable = self.nested_column.nullable
        # Prevent null map reading. Reset nested column nullable flag.
        self.nested_column.nullable = False

        index_size = await self.reader.read_uint64()
        index = await self.nested_column.read_data(index_size,)
        if nullable:
            index = (None,) + index[1:]

        await self.reader.read_uint64()  # number of keys
        keys = await keys_column.read_data(n_items,)

        return tuple(index[x] for x in keys)



import functools

from lz4 import block

from asynch.proto.compression import BaseCompressor, BaseDecompressor
from asynch.proto.protocol import CompressionMethod, CompressionMethodByte


class Compressor(BaseCompressor):
    method = CompressionMethod.LZ4
    method_byte = CompressionMethodByte.LZ4
    compress_func = functools.partial(block.compress, mode="default")


class Decompressor(BaseDecompressor):
    method = CompressionMethod.LZ4
    method_byte = CompressionMethodByte.LZ4
    decompress_func = block.decompress

    async def get_decompressed_data(self, method_byte, compressed_hash, extra_header_size):
        compressed = await super(Decompressor, self).get_decompressed_data(
            method_byte, compressed_hash, extra_header_size
        )

        uncompressed_size = await self.reader.read_uint32()

        return block.decompress(compressed, uncompressed_size=uncompressed_size)



import getpass
import socket

from asynch.errors import LogicalError
from asynch.proto import constants
from asynch.proto.io import BufferedWriter


class ServerInfo:
    def __init__(
        self,
        name: str,
        version_major: int,
        version_minor: int,
        version_patch: int,
        revision: int,
        timezone: str,
        display_name: str,
    ):
        self.name = name
        self.version_major = version_major
        self.version_minor = version_minor
        self.version_patch = version_patch
        self.revision = revision
        self.timezone = timezone
        self.display_name = display_name

    def version_tuple(self):
        return self.version_major, self.version_minor, self.version_patch


class Interface:
    TCP = 1
    HTTP = 2


class QueryKind:
    # Uninitialized object.
    NO_QUERY = 0

    INITIAL_QUERY = 1

    # Query that was initiated by another query for distributed query
    # execution.
    SECONDARY_QUERY = 2


class ClientInfo:
    client_version_major = constants.CLIENT_VERSION_MAJOR
    client_version_minor = constants.CLIENT_VERSION_MINOR
    client_version_patch = constants.CLIENT_VERSION_PATCH
    client_revision = constants.CLIENT_REVISION
    interface = Interface.TCP

    initial_user = ""
    initial_query_id = ""
    initial_address = "0.0.0.0:0"

    quota_key = ""

    def __init__(self, client_name: str):
        self.query_kind = QueryKind.NO_QUERY

        try:
            self.os_user = getpass.getuser()
        except KeyError:
            self.os_user = ""
        self.client_hostname = socket.gethostname()
        self.client_name = client_name

    @property
    def empty(self):
        return self.query_kind == QueryKind.NO_QUERY

    async def write(self, server_revision: int, writer: BufferedWriter):
        revision = server_revision
        if server_revision < constants.DBMS_MIN_REVISION_WITH_CLIENT_INFO:
            raise LogicalError(
                "Method ClientInfo.write is called " "for unsupported server revision"
            )

        await writer.write_int8(self.query_kind,)
        if self.empty:
            return

        await writer.write_str(self.initial_user,)
        await writer.write_str(self.initial_query_id,)
        await writer.write_str(self.initial_address,)

        await writer.write_uint8(self.interface,)

        await writer.write_str(self.os_user,)
        await writer.write_str(self.client_hostname,)
        await writer.write_str(self.client_name,)
        await writer.write_varint(self.client_version_major,)
        await writer.write_varint(self.client_version_minor,)
        await writer.write_varint(self.client_revision,)

        if revision >= constants.DBMS_MIN_REVISION_WITH_QUOTA_KEY_IN_CLIENT_INFO:
            await writer.write_str(self.quota_key,)

        if revision >= constants.DBMS_MIN_REVISION_WITH_VERSION_PATCH:
            await writer.write_varint(self.client_version_patch,)



from .intcolumn import Int64Column


class IntervalColumn(Int64Column):
    pass


class IntervalDayColumn(IntervalColumn):
    ch_type = "IntervalDay"


class IntervalWeekColumn(IntervalColumn):
    ch_type = "IntervalWeek"


class IntervalMonthColumn(IntervalColumn):
    ch_type = "IntervalMonth"


class IntervalYearColumn(IntervalColumn):
    ch_type = "IntervalYear"


class IntervalHourColumn(IntervalColumn):
    ch_type = "IntervalHour"


class IntervalMinuteColumn(IntervalColumn):
    ch_type = "IntervalMinute"


class IntervalSecondColumn(IntervalColumn):
    ch_type = "IntervalSecond"



from asynch.proto.settings.types import (
    SettingBool,
    SettingChar,
    SettingFloat,
    SettingMaxThreads,
    SettingString,
    SettingUInt64,
)

SettingInt64 = SettingUInt64

# Seconds and milliseconds should be set as ints.
SettingSeconds = SettingMilliseconds = SettingUInt64

# Server cares about possible choices validation.
# See https://github.com/yandex/ClickHouse/blob/master/dbms/src/
# Interpreters/Settings.h for all choices.
SettingLoadBalancing = (
    SettingTotalsMode
) = (
    SettingCompressionMethod
) = (
    SettingDistributedProductMode
) = (
    SettingGlobalSubqueriesMethod
) = (
    SettingDateTimeInputFormat
) = (
    SettingURI
) = (
    SettingJoinAlgorithm
) = SettingSpecialSort = SettingLogQueriesType = SettingDefaultDatabaseEngine = SettingString

settings = {
    # Settings
    "min_compress_block_size": SettingUInt64,
    "max_compress_block_size": SettingUInt64,
    "max_block_size": SettingUInt64,
    "max_insert_block_size": SettingUInt64,
    "min_insert_block_size_rows": SettingUInt64,
    "min_insert_block_size_bytes": SettingUInt64,
    "max_partitions_per_insert_block": SettingUInt64,
    "max_threads": SettingMaxThreads,
    "max_alter_threads": SettingMaxThreads,
    "max_read_buffer_size": SettingUInt64,
    "max_distributed_connections": SettingUInt64,
    "max_query_size": SettingUInt64,
    "interactive_delay": SettingUInt64,
    "connect_timeout": SettingSeconds,
    "connect_timeout_with_failover_ms": SettingMilliseconds,
    "receive_timeout": SettingSeconds,
    "send_timeout": SettingSeconds,
    "queue_max_wait_ms": SettingMilliseconds,
    "poll_interval": SettingUInt64,
    "distributed_connections_pool_size": SettingUInt64,
    "connections_with_failover_max_tries": SettingUInt64,
    "extremes": SettingBool,
    "use_uncompressed_cache": SettingBool,
    "replace_running_query": SettingBool,
    "background_pool_size": SettingUInt64,
    "background_schedule_pool_size": SettingUInt64,
    "distributed_directory_monitor_sleep_time_ms": SettingMilliseconds,
    "distributed_directory_monitor_batch_inserts": SettingBool,
    "distributed_directory_monitor_max_sleep_time_ms": SettingMilliseconds,
    "optimize_move_to_prewhere": SettingBool,
    "optimize_skip_unused_shards": SettingBool,
    "optimize_read_in_order": SettingBool,
    "optimize_min_equality_disjunction_chain_length": SettingUInt64,
    "enable_optimize_predicate_expression": SettingBool,
    "replication_alter_partitions_sync": SettingUInt64,
    "replication_alter_columns_timeout": SettingUInt64,
    "load_balancing": SettingLoadBalancing,
    "totals_mode": SettingTotalsMode,
    "totals_auto_threshold": SettingFloat,
    "compile": SettingBool,
    "compile_expressions": SettingBool,
    "min_count_to_compile": SettingUInt64,
    "group_by_two_level_threshold": SettingUInt64,
    "group_by_two_level_threshold_bytes": SettingUInt64,
    "distributed_aggregation_memory_efficient": SettingBool,
    "aggregation_memory_efficient_merge_threads": SettingUInt64,
    "max_parallel_replicas": SettingUInt64,
    "parallel_replicas_count": SettingUInt64,
    "parallel_replica_offset": SettingUInt64,
    "skip_unavailable_shards": SettingBool,
    "distributed_group_by_no_merge": SettingBool,
    "merge_tree_min_rows_for_concurrent_read": SettingUInt64,
    "merge_tree_min_rows_for_seek": SettingUInt64,
    "merge_tree_coarse_index_granularity": SettingUInt64,
    "merge_tree_max_rows_to_use_cache": SettingUInt64,
    "merge_tree_min_bytes_for_concurrent_read": SettingUInt64,
    "merge_tree_min_bytes_for_seek": SettingUInt64,
    "merge_tree_max_bytes_to_use_cache": SettingUInt64,
    "merge_tree_uniform_read_distribution": SettingBool,
    "mysql_max_rows_to_insert": SettingUInt64,
    "min_bytes_to_use_direct_io": SettingUInt64,
    "force_index_by_date": SettingBool,
    "force_primary_key": SettingBool,
    "mark_cache_min_lifetime": SettingUInt64,
    "max_streams_to_max_threads_ratio": SettingFloat,
    "network_compression_method": SettingCompressionMethod,
    "network_zstd_compression_level": SettingInt64,
    "priority": SettingUInt64,
    "log_queries": SettingBool,
    "log_queries_cut_to_length": SettingUInt64,
    "query_profiler_real_time_period_ns": SettingUInt64,
    "query_profiler_cpu_time_period_ns": SettingUInt64,
    "enable_debug_queries": SettingBool,
    "distributed_product_mode": SettingDistributedProductMode,
    "max_concurrent_queries_for_user": SettingUInt64,
    "insert_deduplicate": SettingBool,
    "insert_quorum": SettingUInt64,
    "insert_quorum_timeout": SettingMilliseconds,
    "select_sequential_consistency": SettingUInt64,
    "table_function_remote_max_addresses": SettingUInt64,
    "read_backoff_min_latency_ms": SettingMilliseconds,
    "read_backoff_max_throughput": SettingUInt64,
    "read_backoff_min_interval_between_events_ms": SettingMilliseconds,
    "read_backoff_min_events": SettingUInt64,
    "memory_tracker_fault_probability": SettingFloat,
    "enable_http_compression": SettingBool,
    "http_zlib_compression_level": SettingInt64,
    "http_native_compression_disable_checksumming_on_decompress": SettingBool,
    "count_distinct_implementation": SettingString,
    "add_http_cors_header": SettingBool,
    "input_format_skip_unknown_fields": SettingBool,
    "input_format_import_nested_json": SettingBool,
    "input_format_values_interpret_expressions": SettingBool,
    "input_format_with_names_use_header": SettingBool,
    "input_format_defaults_for_omitted_fields": SettingBool,
    "input_format_null_as_default": SettingBool,
    "input_format_values_deduce_templates_of_expressions": SettingBool,
    "input_format_values_accurate_types_of_literals": SettingBool,
    "input_format_allow_errors_num": SettingUInt64,
    "input_format_allow_errors_ratio": SettingFloat,
    "input_format_csv_unquoted_null_literal_as_null": SettingBool,
    "output_format_write_statistics": SettingBool,
    "output_format_json_quote_64bit_integers": SettingBool,
    "output_format_json_quote_denormals": SettingBool,
    "output_format_json_escape_forward_slashes": SettingBool,
    "output_format_pretty_max_rows": SettingUInt64,
    "output_format_pretty_max_column_pad_width": SettingUInt64,
    "output_format_pretty_color": SettingBool,
    "output_format_parquet_row_group_size": SettingUInt64,
    "use_client_time_zone": SettingBool,
    "send_progress_in_http_headers": SettingBool,
    "http_headers_progress_interval_ms": SettingUInt64,
    "fsync_metadata": SettingBool,
    "join_use_nulls": SettingBool,
    "join_default_strictness": SettingString,
    "preferred_block_size_bytes": SettingUInt64,
    "max_replica_delay_for_distributed_queries": SettingUInt64,
    "fallback_to_stale_replicas_for_distributed_queries": SettingBool,
    "preferred_max_column_in_block_size_bytes": SettingUInt64,
    "insert_distributed_sync": SettingBool,
    "insert_distributed_timeout": SettingUInt64,
    "distributed_ddl_task_timeout": SettingInt64,
    "stream_flush_interval_ms": SettingMilliseconds,
    "format_schema": SettingString,
    "insert_allow_materialized_columns": SettingBool,
    "http_connection_timeout": SettingSeconds,
    "http_send_timeout": SettingSeconds,
    "http_receive_timeout": SettingSeconds,
    "optimize_throw_if_noop": SettingBool,
    "use_index_for_in_with_subqueries": SettingBool,
    "empty_result_for_aggregation_by_empty_set": SettingBool,
    "allow_distributed_ddl": SettingBool,
    "odbc_max_field_size": SettingUInt64,
    # Limits
    "max_rows_to_read": SettingUInt64,
    "max_bytes_to_read": SettingUInt64,
    "read_overflow_mode": SettingString,
    "max_rows_to_group_by": SettingUInt64,
    "group_by_overflow_mode": SettingString,
    "max_bytes_before_external_group_by": SettingUInt64,
    "max_rows_to_sort": SettingUInt64,
    "max_bytes_to_sort": SettingUInt64,
    "sort_overflow_mode": SettingString,
    "max_bytes_before_external_sort": SettingUInt64,
    "max_bytes_before_remerge_sort": SettingUInt64,
    "max_result_rows": SettingUInt64,
    "max_result_bytes": SettingUInt64,
    "result_overflow_mode": SettingString,
    "max_execution_time": SettingSeconds,
    "timeout_overflow_mode": SettingString,
    "min_execution_speed": SettingUInt64,
    "timeout_before_checking_execution_speed": SettingSeconds,
    "max_columns_to_read": SettingUInt64,
    "max_temporary_columns": SettingUInt64,
    "max_temporary_non_const_columns": SettingUInt64,
    "max_subquery_depth": SettingUInt64,
    "max_pipeline_depth": SettingUInt64,
    "max_ast_depth": SettingUInt64,
    "max_ast_elements": SettingUInt64,
    "max_expanded_ast_elements": SettingUInt64,
    "readonly": SettingUInt64,
    "max_rows_in_set": SettingUInt64,
    "max_bytes_in_set": SettingUInt64,
    "set_overflow_mode": SettingString,
    "max_rows_in_join": SettingUInt64,
    "max_bytes_in_join": SettingUInt64,
    "join_overflow_mode": SettingString,
    "max_rows_to_transfer": SettingUInt64,
    "max_bytes_to_transfer": SettingUInt64,
    "transfer_overflow_mode": SettingString,
    "max_rows_in_distinct": SettingUInt64,
    "max_bytes_in_distinct": SettingUInt64,
    "distinct_overflow_mode": SettingString,
    "max_memory_usage": SettingUInt64,
    "max_memory_usage_for_user": SettingUInt64,
    "max_memory_usage_for_all_queries": SettingUInt64,
    "max_network_bandwidth": SettingUInt64,
    "max_network_bytes": SettingUInt64,
    "max_network_bandwidth_for_user": SettingUInt64,
    "max_network_bandwidth_for_all_users": SettingUInt64,
    "max_streams_multiplier_for_merge_tables": SettingFloat,
    "max_http_get_redirects": SettingUInt64,
    "max_execution_speed": SettingUInt64,
    "max_execution_speed_bytes": SettingUInt64,
    "format_csv_delimiter": SettingChar,
    "format_csv_allow_single_quotes": SettingBool,
    "format_csv_allow_double_quotes": SettingBool,
    "format_template_resultset": SettingString,
    "format_template_row": SettingString,
    "format_template_rows_between_delimiter": SettingString,
    "format_custom_escaping_rule": SettingString,
    "format_custom_field_delimiter": SettingString,
    "format_custom_row_before_delimiter": SettingString,
    "format_custom_row_after_delimiter": SettingString,
    "format_custom_row_between_delimiter": SettingString,
    "format_custom_result_before_delimiter": SettingString,
    "format_custom_result_after_delimiter": SettingString,
    "enable_conditional_computation": SettingUInt64,
    "date_time_input_format": SettingDateTimeInputFormat,
    "log_profile_events": SettingBool,
    "log_query_settings": SettingBool,
    "log_query_threads": SettingBool,
    "send_logs_level": SettingString,
    "low_cardinality_max_dictionary_size": SettingUInt64,
    "low_cardinality_use_single_dictionary_for_part": SettingBool,
    "decimal_check_overflow": SettingBool,
    "prefer_localhost_replica": SettingBool,
    "max_fetch_partition_retries_count": SettingUInt64,
    "asterisk_left_columns_only": SettingBool,
    "http_max_multipart_form_data_size": SettingUInt64,
    "calculate_text_stack_trace": SettingBool,
    "parallel_view_processing": SettingBool,
    "allow_experimental_low_cardinality_type": SettingBool,
    "allow_experimental_decimal_type": SettingBool,
    "allow_suspicious_low_cardinality_types": SettingBool,
    "allow_experimental_multiple_joins_emulation": SettingBool,
    "allow_experimental_cross_to_join_conversion": SettingBool,
    "allow_experimental_data_skipping_indices": SettingBool,
    "allow_hyperscan": SettingBool,
    "allow_simdjson": SettingBool,
    "allow_introspection_functions": SettingBool,
    "allow_drop_detached": SettingBool,
    "allow_experimental_live_view": SettingBool,
    "allow_ddl": SettingBool,
    "partial_merge_join": SettingBool,
    "partial_merge_join_optimizations": SettingBool,
    "partial_merge_join_rows_in_right_blocks": SettingUInt64,
    "partial_merge_join_rows_in_left_blocks": SettingFloat,
    "distributed_replica_error_half_life": SettingSeconds,
    "distributed_replica_error_cap": SettingUInt64,
    "min_free_disk_space_for_temporary_data": SettingUInt64,
    "tcp_keep_alive_timeout": SettingSeconds,
    "connection_pool_max_wait_ms": SettingMilliseconds,
    "kafka_max_wait_ms": SettingMilliseconds,
    "idle_connection_timeout": SettingUInt64,
    "s3_min_upload_part_size": SettingUInt64,
    "any_join_distinct_right_table_keys": SettingBool,
    "join_any_take_last_row": SettingBool,
    "stream_poll_timeout_ms": SettingMilliseconds,
    "joined_subquery_requires_alias": SettingBool,
    "enable_unaligned_array_join": SettingBool,
    "low_cardinality_allow_in_native_format": SettingBool,
    "external_table_functions_use_nulls": SettingBool,
    "experimental_use_processors": SettingBool,
    "check_query_single_value_result": SettingBool,
    "live_view_heartbeat_interval": SettingSeconds,
    "temporary_live_view_timeout": SettingSeconds,
    "max_live_view_insert_blocks_before_refresh": SettingUInt64,
    "max_insert_threads": SettingUInt64,
    "replace_running_query_max_wait_ms": SettingMilliseconds,
    "background_move_pool_size": SettingUInt64,
    "min_count_to_compile_expression": SettingUInt64,
    "force_optimize_skip_unused_shards": SettingUInt64,
    "input_format_parallel_parsing": SettingBool,
    "min_chunk_bytes_for_parallel_parsing": SettingUInt64,
    "min_bytes_to_use_mmap_io": SettingUInt64,
    "os_thread_priority": SettingInt64,
    "input_format_tsv_empty_as_default": SettingBool,
    "input_format_avro_schema_registry_url": SettingString,
    "output_format_avro_codec": SettingString,
    "output_format_avro_sync_interval": SettingUInt64,
    "min_execution_speed_bytes": SettingUInt64,
    "default_max_bytes_in_join": SettingUInt64,
    "enable_optimize_predicate_expression_to_final_subquery": SettingBool,
    "cancel_http_readonly_queries_on_client_close": SettingBool,
    "enable_scalar_subquery_optimization": SettingBool,
    "optimize_trivial_count_query": SettingBool,
    "mutations_sync": SettingUInt64,
    "optimize_if_chain_to_miltiif": SettingBool,
    "max_parser_depth": SettingUInt64,
    "max_joined_block_size_rows": SettingUInt64,
    "connect_timeout_with_failover_secure_ms": SettingMilliseconds,
    "parallel_distributed_insert_select": SettingBool,
    "force_optimize_skip_unused_shards_no_nested": SettingBool,
    "format_avro_schema_registry_url": SettingURI,
    "output_format_tsv_crlf_end_of_line": SettingBool,
    "join_algorithm": SettingJoinAlgorithm,
    "memory_profiler_step": SettingUInt64,
    "output_format_csv_crlf_end_of_line": SettingBool,
    "allow_experimental_alter_materialized_view_structure": SettingBool,
    "enable_early_constant_folding": SettingBool,
    "deduplicate_blocks_in_dependent_materialized_views": SettingBool,
    "use_compact_format_in_distributed_parts_names": SettingBool,
    "multiple_joins_rewriter_version": SettingUInt64,
    "min_insert_block_size_rows_for_materialized_views": SettingUInt64,
    "min_insert_block_size_bytes_for_materialized_views": SettingUInt64,
    "max_final_threads": SettingUInt64,
    "background_buffer_flush_schedule_pool_size": SettingUInt64,
    "background_distributed_schedule_pool_size": SettingUInt64,
    "special_sort": SettingSpecialSort,
    "optimize_distributed_group_by_sharding_key": SettingBool,
    "log_queries_min_type": SettingLogQueriesType,
    "allow_suspicious_codecs": SettingBool,
    "metrics_perf_events_enabled": SettingBool,
    "metrics_perf_events_list": SettingString,
    "join_on_disk_max_files_to_merge": SettingUInt64,
    "temporary_files_codec": SettingString,
    "max_untracked_memory": SettingUInt64,
    "memory_profiler_sample_probability": SettingFloat,
    "optimize_aggregation_in_order": SettingBool,
    "default_database_engine": SettingDefaultDatabaseEngine,
    "allow_experimental_database_atomic": SettingBool,
    "show_table_uuid_in_table_create_query_if_not_nil": SettingBool,
    "optimize_arithmetic_operations_in_aggregate_functions": SettingBool,
    "validate_polygons": SettingBool,
    "transform_null_in": SettingBool,
    "allow_nondeterministic_mutations": SettingBool,
    "lock_acquire_timeout": SettingSeconds,
    "materialize_ttl_after_modify": SettingBool,
    "allow_experimental_geo_types": SettingBool,
    "output_format_pretty_max_value_width": SettingUInt64,
    "format_regexp": SettingString,
    "format_regexp_escaping_rule": SettingString,
    "format_regexp_skip_unmatched": SettingBool,
    "output_format_enable_streaming": SettingBool,
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asynch',
 'asynch.proto',
 'asynch.proto.columns',
 'asynch.proto.compression',
 'asynch.proto.settings',
 'asynch.proto.streams',
 'asynch.proto.utils']

package_data = \
{'': ['*']}

install_requires = \
['ciso8601', 'clickhouse-cityhash', 'leb128', 'lz4', 'pytz', 'tzlocal', 'zstd']

setup_kwargs = {
    'name': 'asynch',
    'version': '0.1.7',
    'description': 'A asyncio driver for ClickHouse with native tcp protocol',
    'long_description': '# asynch\n\n![pypi](https://img.shields.io/pypi/v/asynch.svg?style=flat)\n![license](https://img.shields.io/github/license/long2ice/asynch)\n![workflows](https://github.com/long2ice/asynch/workflows/pypi/badge.svg)\n![workflows](https://github.com/long2ice/asynch/workflows/ci/badge.svg)\n\n## Introduction\n\n`asynch` is an asyncio ClickHouse Python Driver with native (TCP) interface support, which reuse most of [clickhouse-driver](https://github.com/mymarilyn/clickhouse-driver) and comply with [PEP249](https://www.python.org/dev/peps/pep-0249/).\n\n## Install\n\n```shell\n> pip install asynch\n```\n\n## Usage\n\nConnect to ClickHouse\n\n```python\nfrom asynch import connect\n\nasync def connect_database():\n    conn = await connect(\n        host = "127.0.0.1",\n        port = 9000,\n        database = "default",\n        user = "default",\n        password = "",\n    )\n```\n\nCreate table by sql\n\n```python\nasync def create_table():\n    async with conn.cursor(cursor=DictCursor) as cursor:\n        await cursor.execute(\'create database if not exists test\')\n        await cursor.execute("""\n        CREATE TABLE if not exists test.asynch\n            (\n                `id`       Int32,\n                `decimal`  Decimal(10, 2),\n                `date`     Date,\n                `datetime` DateTime,\n                `float`    Float32,\n                `uuid`     UUID,\n                `string`   String,\n                `ipv4`     IPv4,\n                `ipv6`     IPv6\n\n            )\n            ENGINE = MergeTree\n                ORDER BY id"""\n        )\n```\n\nUse `fetchone`\n\n```python\nasync def fetchone():\n    async with conn.cursor() as cursor:\n        await cursor.execute("SELECT 1")\n        ret = cursor.fetchone()\n        assert ret == (1,)\n```\n\nUse `fetchmany`\n\n```python\nasync def fetchall():\n    async with conn.cursor() as cursor:\n        await cursor.execute("SELECT 1")\n        ret = cursor.fetchall()\n        assert ret == [(1,)]\n```\n\nUse `DictCursor` to get result with dict\n\n```python\nasync def dict_cursor():\n    async with conn.cursor(cursor=DictCursor) as cursor:\n        await cursor.execute("SELECT 1")\n        ret = cursor.fetchall()\n        assert ret == [{"1": 1}]\n```\n\nInsert data with dict\n\n```python\nfrom asynch.cursors import DictCursor\n\nasync def insert_dict():\n    async with conn.cursor(cursor=DictCursor) as cursor:\n        ret = await cursor.execute(\n            """INSERT INTO test.asynch(id,decimal,date,datetime,float,uuid,string,ipv4,ipv6) VALUES""",\n            [\n                {\n                    "id": 1,\n                    "decimal": 1,\n                    "date": "2020-08-08",\n                    "datetime": "2020-08-08 00:00:00",\n                    "float": 1,\n                    "uuid": "59e182c4-545d-4f30-8b32-cefea2d0d5ba",\n                    "string": "1",\n                    "ipv4": "0.0.0.0",\n                    "ipv6": "::",\n                }\n            ],\n        )\n        assert ret == 1\n```\n\nInsert data with tuple\n\n```python\nasync def insert_tuple():\n    async with conn.cursor(cursor=DictCursor) as cursor:\n        ret = await cursor.execute(\n            """INSERT INTO test.asynch(id,decimal,date,datetime,float,uuid,string,ipv4,ipv6) VALUES""",\n            [\n                (\n                    1,\n                    1,\n                    "2020-08-08",\n                    "2020-08-08 00:00:00",\n                    1,\n                    "59e182c4-545d-4f30-8b32-cefea2d0d5ba",\n                    "1",\n                    "0.0.0.0",\n                    "::",\n                )\n            ],\n        )\n        assert ret == 1\n```\n\nUse connection pool\n\n```python\nasync def use_pool():\n    pool = await asynch.create_pool()\n    async with pool.acquire() as conn:\n        async with conn.cursor() as cursor:\n            await cursor.execute("SELECT 1")\n            ret = cursor.fetchone()\n            assert ret == (1,)\n    pool.close()\n    await pool.wait_closed()\n```\n\n## ThanksTo\n\n- [clickhouse-driver](https://github.com/mymarilyn/clickhouse-driver), ClickHouse Python Driver with native interface support.\n\n## License\n\nThis project is licensed under the [Apache-2.0](https://github.com/long2ice/asynch/blob/master/LICENSE) License.\n',
    'author': 'long2ice',
    'author_email': 'long2ice@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/long2ice/asynch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from clickhouse_cityhash.cityhash import CityHash128

from asynch.proto.compression import get_decompressor_cls
from asynch.proto.context import Context
from asynch.proto.io import BufferedReader, BufferedWriter
from asynch.proto.streams.native import BlockInputStream, BlockOutputStream


class CompressedBlockOutputStream(BlockOutputStream):
    def __init__(
        self,
        reader: BufferedReader,
        writer: BufferedWriter,
        context: Context,
        compressor_cls,
        compress_block_size,
    ):
        super().__init__(reader, writer, context)
        self.compressor_cls = compressor_cls
        self.compress_block_size = compress_block_size

        self.compressor = self.compressor_cls(writer)

    def get_compressed_hash(self, data):
        return CityHash128(data)

    async def finalize(self):
        await self.writer.flush()

        compressed = await self.get_compressed()
        compressed_size = len(compressed)

        compressed_hash = self.get_compressed_hash(compressed)
        await self.writer.write_uint128(compressed_hash,)

        block_size = self.compress_block_size

        i = 0
        while i < compressed_size:
            await self.writer.write_bytes(compressed[i : i + block_size])  # noqa: E203
            i += block_size

        await self.writer.flush()

    async def get_compressed(self):
        compressed = BufferedWriter()

        if self.compressor.method_byte is not None:
            await compressed.write_uint8(self.compressor.method_byte)
            extra_header_size = 1  # method
        else:
            extra_header_size = 0

        data = self.compressor.get_compressed_data(extra_header_size)
        await compressed.write_bytes(data)

        return compressed.buffer


class CompressedBlockInputStream(BlockInputStream):
    def get_compressed_hash(self, data):
        return CityHash128(data)

    async def read_block(self):
        compressed_hash = await self.reader.read_uint128()
        method_byte = await self.reader.read_uint8()

        decompressor_cls = get_decompressor_cls(method_byte)
        decompressor = decompressor_cls(self.reader)

        if decompressor.method_byte is not None:
            extra_header_size = 1  # method
        else:
            extra_header_size = 0

        return decompressor.get_decompressed_data(method_byte, compressed_hash, extra_header_size)



import struct
from asyncio import StreamReader, StreamWriter

import leb128

MAX_UINT64 = (1 << 64) - 1
MAX_INT64 = (1 << 63) - 1


class BufferedWriter:
    def __init__(self, writer: StreamWriter = None, max_buffer_size: int = 0):
        self.max_buffer_size = max_buffer_size
        self.writer = writer
        self.buffer = bytearray()
        self.position = 0

    async def flush(self):
        if not self.writer:
            return
        self.writer.write(self.buffer)
        self.buffer = bytearray()
        self.position = 0
        await self.writer.drain()

    async def write_bytes(self, data: bytes):
        self.buffer.extend(data)
        self.position += len(data)
        if self.position >= self.max_buffer_size:
            await self.flush()

    async def write_varint(self, data: int):
        packet = leb128.i.encode(data)
        await self.write_bytes(packet)

    async def write_str(self, data: str):
        packet = data.encode()
        await self.write_varint(len(packet))
        await self.write_bytes(packet)

    async def write_strings(self, data):
        for item in data:
            if isinstance(item, str):
                packet = item.encode()
            else:
                packet = item
            await self.write_varint(len(packet))
            await self.write_bytes(packet)

    async def close(self):
        if not self.writer:
            return
        self.writer.close()
        await self.writer.wait_closed()

    async def write_int(self, data: int, fmt: str):
        fmt = "<" + fmt
        await self.write_bytes(struct.pack(fmt, data))

    async def write_int8(
        self, data: int,
    ):
        await self.write_int(data, "b")

    async def write_int16(self, data):
        await self.write_int(data, "h")

    async def write_int32(self, data):
        await self.write_int(data, "i")

    async def write_int64(self, data):
        await self.write_int(data, "q")

    async def write_uint8(self, data):
        await self.write_int(data, "B")

    async def write_uint16(self, data):
        await self.write_int(data, "H")

    async def write_uint32(
        self, data: int,
    ):
        await self.write_int(data, "I")

    async def write_uint64(self, data: int):
        await self.write_int(data, "Q")

    async def write_uint128(self, data: int):
        fmt = "<QQ"
        packet = struct.pack(fmt, (data >> 64) & MAX_UINT64, data & MAX_UINT64)
        await self.write_bytes(packet)


class BufferedReader:
    def __init__(self, reader: StreamReader, buffer_max_size: int):
        self.buffer_max_size = buffer_max_size
        self.reader = reader
        self.buffer = bytearray()
        self.current_buffer_size = 0
        self.position = 0

    async def _read_into_buffer(self):
        packet = await self.reader.read(self.buffer_max_size)
        self.buffer.extend(packet)
        self.current_buffer_size = len(self.buffer)

    def _read_one(self):
        packet = self.buffer[self.position]
        self.position += 1
        return packet

    async def read_varint(self):
        if self.position == self.current_buffer_size:
            self._reset_buffer()
            await self._read_into_buffer()
        packets = bytearray()
        while True:
            packet = self._read_one()
            packets.append(packet)
            if packet < 0x80:
                break
        return leb128.u.decode(packets)

    def _reset_buffer(self):
        self.position = 0
        self.buffer = bytearray()

    async def read_str(self):
        length = await self.read_varint()
        packet = await self.read_bytes(length)
        return packet.decode()

    async def read_bytes(self, length: int):
        packets = []
        while length > 0:
            if self.position == self.current_buffer_size:
                self._reset_buffer()
                await self._read_into_buffer()

            read_position = self.position + length
            packet = self.buffer[self.position : read_position]  # noqa: E203
            length -= len(packet)
            self.position += len(packet)
            packets.append(packet)

        return b"".join(packets)

    async def read_int(self, fmt: str):
        s = struct.Struct("<" + fmt)
        packet = await self.read_bytes(s.size)
        return s.unpack(packet)[0]

    async def read_int8(self,):
        return await self.read_int("b")

    async def read_int16(self,):
        return await self.read_int("h")

    async def read_int32(self,):
        return await self.read_int("i")

    async def read_int64(self,):
        return await self.read_int("q")

    async def read_uint8(self,):
        return await self.read_int("B")

    async def read_uint16(self,):
        return await self.read_int("H")

    async def read_uint32(self,):
        return await self.read_int("I")

    async def read_uint64(self,):
        return await self.read_int("Q")

    async def read_uint128(self,):
        hi = await self.read_int("Q")
        lo = await self.read_int("Q")
        return (hi << 64) + lo



from asynch.proto import constants
from asynch.proto.block import BaseBlock, BlockInfo, ColumnOrientedBlock
from asynch.proto.columns import read_column, write_column
from asynch.proto.context import Context
from asynch.proto.io import BufferedReader, BufferedWriter


class BlockOutputStream:
    def __init__(self, reader: BufferedReader, writer: BufferedWriter, context: Context):
        self.reader = reader
        self.writer = writer
        self.context = context

    async def write(self, block: BaseBlock):
        revision = self.context.server_info.revision
        if revision >= constants.DBMS_MIN_REVISION_WITH_BLOCK_INFO:
            await block.info.write()

        # We write transposed data.
        n_columns = block.num_columns
        n_rows = block.num_rows

        await self.writer.write_varint(n_columns)
        await self.writer.write_varint(n_rows)

        for i, (col_name, col_type) in enumerate(block.columns_with_types):
            await self.writer.write_str(col_name,)
            await self.writer.write_str(col_type,)

            if n_columns:
                try:
                    items = block.get_column_by_index(i)
                except IndexError:
                    raise ValueError("Different rows length")

                await write_column(
                    self.reader,
                    self.writer,
                    self.context,
                    col_name,
                    col_type,
                    items,
                    types_check=block.types_check,
                )

        await self.finalize()

    async def finalize(self):
        await self.writer.flush()


class BlockInputStream:
    def __init__(self, reader: BufferedReader, writer: BufferedWriter, context):
        self.writer = writer
        self.reader = reader
        self.context = context

    async def read(self):
        info = BlockInfo(reader=self.reader)

        revision = self.context.server_info.revision
        if revision >= constants.DBMS_MIN_REVISION_WITH_BLOCK_INFO:
            await info.read()

        n_columns = await self.reader.read_varint()
        n_rows = await self.reader.read_varint()

        data, names, types = [], [], []

        for i in range(n_columns):
            column_name = await self.reader.read_str()
            column_type = await self.reader.read_str()

            names.append(column_name)
            types.append(column_type)

            if n_rows:
                column = await read_column(
                    self.reader, self.writer, self.context, column_type, n_rows,
                )
                data.append(column)

        block = ColumnOrientedBlock(
            reader=self.reader, columns_with_types=list(zip(names, types)), data=data, info=info,
        )

        return block
